@echo off
REM -----------------------------------------------
REM Script to customize module-tcl for "cmd" shell
REM Load the modulepaths with "module use"
REM -----------------------------------------------
call module use %MODULESHOME%\modulefiles
call module avail
